import io
import PIL
import numpy as np
import matplotlib.pyplot as plt

from pyboreas.data.bounding_boxes import BoundingBox2D
from pyboreas.utils.utils import rotToYawPitchRoll, yaw


def convert_plt_to_img(dpi=128):
    buf = io.BytesIO()
    plt.savefig(buf, format='png', dpi=dpi, bbox_inches='tight', pad_inches=0)
    plt.close()
    buf.seek(0)
    return PIL.Image.open(buf)


def transform_bounding_boxes(T, C_yaw, raw_labels):
    """
    Generate bounding boxes from labels and transform them
    by a SE3 transformation
    :param T: required SE3 transformation
    :param C_yaw: yaw component of the SE3 transformation
    :param raw_labels: original label data
    """
    boxes = []
    for i in range(len(raw_labels)):
        # Load Labels
        bbox_raw_pos = np.concatenate(
            (np.fromiter(raw_labels[i]['position'].values(), dtype=float), [1]))
        # Create Bounding Box
        pos = np.matmul(T, np.array([bbox_raw_pos]).T)[:3]
        rotation = np.matmul(C_yaw, yaw(raw_labels[i]['yaw']))
        rotToYawPitchRoll(rotation)
        extent = np.array(list(raw_labels[i]['dimensions'].values())).reshape(3, 1)  # Convert to 2d
        box = BoundingBox2D(pos, rotation, extent, raw_labels[i]['label'])
        boxes.append(box)
    return boxes


def vis_camera(cam, figsize=(24.48, 20.48), dpi=100):
    fig = plt.figure(figsize=figsize, dpi=dpi)
    ax = fig.add_subplot()
    ax.imshow(cam.img[:, :, ::-1])
    ax.set_axis_off()
    plt.show()


def vis_lidar(lid, bounds=[-40, 40, -40, 40, -10, 30], figsize=(10, 10), cmap='winter',
              color='intensity', vmin=None, vmax=None, azim_delta=-75, elev_delta=-5):
    lid.passthrough(bounds)
    p = lid.points
    if color == 'x':
        c = p[:, 0]
    elif color == 'y':
        c = p[:, 1]
    elif color == 'z':
        c = p[:, 2]
    elif color == 'intensity':
        c = p[:, 3]
    elif color == 'ring':
        c = p[:, 4]
    elif color == 'time':
        c = p[:, 5]
    else:
        print('warning: color: {} is not valid'.format(color))
        c = p[:, 2]
    if vmin is None or vmax is None:
        vmin = np.min(c)
        vmax = np.max(c)
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(projection='3d')
    ax.azim += azim_delta
    ax.elev += elev_delta
    xs = p[:, 0]
    ys = p[:, 1]
    zs = p[:, 2]
    ax.set_box_aspect((np.ptp(xs), np.ptp(ys), np.ptp(zs)))
    ax.scatter(xs=xs, ys=ys, zs=zs, s=0.1, c=c, cmap=cmap,
               vmin=vmin, vmax=vmax, depthshade=False)
    plt.show()


def vis_radar(rad, figsize=(10, 10), dpi=100, cart_resolution=0.2384, cart_pixel_width=640, cmap='gray'):
    cart = rad.get_cartesian(cart_resolution=cart_resolution, cart_pixel_width=cart_pixel_width, in_place=False)
    fig = plt.figure(figsize=figsize, dpi=dpi)
    ax = fig.add_subplot()
    ax.imshow(cart, cmap=cmap)
    ax.set_axis_off()
    plt.show()



import os
import os.path as osp

from pyboreas.data.splits import *
from pyboreas.data.sequence import Sequence
from pyboreas.data.sensors import Camera, Lidar, Radar
from pyboreas.vis.visualizer import BoreasVisualizer


class BoreasDataset:

    def __init__(self, root='/data/boreas/', split=None, verbose=False):
        self.root = root
        self.split = split
        self.camera_frames = []
        self.lidar_frames = []
        self.radar_frames = []
        self.sequences = []
        self.seqDict = {}  # seq string to index
        self.map = None  # TODO: Load the HD map data

        if split is None:
            split = sorted([[f] for f in os.listdir(root) if 'boreas' in f])

        for seqSpec in split:
            seq = Sequence(root, seqSpec, verbose)
            self.sequences.append(seq)
            self.camera_frames += seq.camera_frames
            self.lidar_frames += seq.lidar_frames
            self.radar_frames += seq.radar_frames
            self.seqDict[seq.ID] = len(self.sequences) - 1

        if verbose:
            print('total camera frames: {}'.format(len(self.camera_frames)))
            print('total lidar frames: {}'.format(len(self.lidar_frames)))
            print('total radar frames: {}'.format(len(self.radar_frames)))

    def get_seq_from_ID(self, ID):
        return self.sequences[self.seqDict[ID]]

    def get_seq(self, idx):
        return self.sequences[idx]

    def get_camera(self, idx):
        self.camera_frames[idx].load_data()
        return self.camera_frames[idx]

    def get_lidar(self, idx):
        self.lidar_frames[idx].load_data()
        return self.lidar_frames[idx]

    def get_radar(self, idx):
        self.radar_frames[idx].load_data()
        return self.radar_frames[idx]

    def get_sequence_visualizer(self, ID):
        return BoreasVisualizer(self.get_seq_from_ID(ID))

# TODO: projectBoundingBoxesOntoSensor(BB, Sensor) (include camera projection?)
# TODO: projectMapOntoView(position, orientation, extent)
# TODO: convert a list of poses to benchmark format
# TODO: provide example code for interpolating between poses (pysteam)



from threading import Lock
import copy

import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.widgets import Button
import numpy as np
import cv2

from pyboreas.vis import map_utils


class BoreasPlotter:
    """
    Class for plotting persp and BEV for boreas stuff
    """
    def __init__(self, visualizer, curr_ts_idx, labels=None, mode='both'):
        # Data
        self.calib = visualizer.calib
        self.lidar_frames = visualizer.lidar_frames
        self.camera_frames = visualizer.camera_frames
        self.radar_frames = visualizer.radar_frames
        
        assert not (self.lidar_frames is None and self.camera_frames is None)  # We must have something to plot

        # For plotting
        self.mode = mode
        self.curr_ts_idx = curr_ts_idx
        self.plot_update_mutex = Lock()

        # Create plot based on mode
        if mode == 'both':
            self.fig, self.ax = plt.subplots(nrows=1, ncols=2, figsize=(20, 7))
            self.bev_ax = self.ax[0]
            self.bev_ax.set_aspect('equal', adjustable='box')
            self.persp_ax = self.ax[1]
        elif mode == 'persp':
            self.fig, self.ax = plt.subplots(nrows=1, ncols=1, figsize=(7, 7))
            self.bev_ax = None
            self.persp_ax = self.ax
        elif mode == 'bev':
            self.fig, self.ax = plt.subplots(nrows=1, ncols=1, figsize=(7, 7))
            self.bev_ax = self.ax
            self.bev_ax.set_aspect('equal', adjustable='box')
            self.persp_ax = None

        # Make buttons for forward/reverse
        button_ax = plt.axes([0.05, 0.05, 0.05, 0.05])
        self.button_f = Button(button_ax, "<")  # Store buttons ref so they dont go out of scope and become useless
        self.button_f.on_clicked(self.on_click_bkwd)

        button_ax2 = plt.axes([0.90, 0.05, 0.05, 0.05])
        self.button_f2 = Button(button_ax2, ">")
        self.button_f2.on_clicked(self.on_click_fwd)

    def on_click_fwd(self, event):
        if not self.plot_update_mutex.acquire(timeout=0.5):
            return
        try:
            self.curr_ts_idx = min(self.curr_ts_idx + 1, len(self.lidar_frames) - 1)
            print("Visualizing Timestep Index: {}/{}...".format(self.curr_ts_idx, len(self.lidar_frames)), end=" ")
            self.update_plots(self.curr_ts_idx)
            print("Done")
        finally:
            self.plot_update_mutex.release()

    def on_click_bkwd(self, event):
        if not self.plot_update_mutex.acquire(timeout=0.5):
            return
        try:
            self.curr_ts_idx = max(self.curr_ts_idx - 1, 0)
            print("Visualizing Timestep Index: {}/{}...".format(self.curr_ts_idx, len(self.lidar_frames)), end=" ")
            self.update_plots(self.curr_ts_idx)
            print("Done")
        finally:
            self.plot_update_mutex.release()

    def update_plots(self, frame_idx):
        if self.mode == 'bev' or self.mode == 'both':
            lidar_frame = self.lidar_frames[frame_idx]
            self.bev_ax.clear()
            self.update_plot_bev(self.bev_ax, lidar_frame)

        if self.mode == 'persp' or self.mode == 'both':
            camera_frame = self.camera_frames[frame_idx]
            if self.lidar_frames is not None:
                lidar_frame = self.lidar_frames[frame_idx]
            else:
                lidar_frame = None
            self.persp_ax.clear()
            self.update_plot_persp(self.persp_ax, camera_frame, lidar_frame)

    def update_plot_bev(self, ax, lidar_frame, downsample_factor=0.3):
        # Calculate transformations for current data
        C_l_enu = lidar_frame.pose[:3, :3].T
        C_a_enu = self.calib.T_applanix_lidar[0:3, 0:3] @ C_l_enu
        C_av = self.calib.T_applanix_lidar[0:3, 0:3]
        lidar_points = lidar_frame.load_data()

        # Draw map CURRENTLY HARDCODED
        map_utils.draw_map("./python/vis/boreas_lane.osm", ax, lidar_frame.pose[0, 3], lidar_frame.pose[1, 3], C_a_enu, utm=True)

        # Calculate point colors
        z_min = -3
        z_max = 5
        colors = cm.jet(((lidar_points[:, 2] - z_min) / (z_max - z_min)) + 0.2, 1)[:, 0:3]

        # Draw lidar points
        pcd_i = np.matmul(C_av[0:2, 0:2].reshape(1, 2, 2), lidar_points[:, 0:2].reshape(lidar_points.shape[0], 2, 1)).squeeze(-1)
        rand_idx = np.random.choice(pcd_i.shape[0], size=int(pcd_i.shape[0]*downsample_factor), replace=False)
        self.scatter = ax.scatter(pcd_i[rand_idx, 0], pcd_i[rand_idx, 1], color=colors[rand_idx, :], s=0.05)

        # Draw predictions (TODO)
        # for box in boxes:
        #     box.render_bbox_2d(ax)
        #
        # if predictions is not None:
        #     for box in predictions:
        #         box.render_bbox_2d(ax, color="k")

        # Set to scale labeling bounds
        ax.set_xlim(-75, 75)
        ax.set_ylim(-75, 75)

        plt.draw()

    def update_plot_persp(self, ax, camera_data, lidar_frame, downsample_factor=0.3):
        if lidar_frame is None:  # If there is no lidar points, just show image
            ax.imshow(camera_data.load_data())
        else:
            points = lidar_frame.load_data()[:, 0:3]
            points = points[np.random.choice(points.shape[0], size=int(points.shape[0]*downsample_factor), replace=False)]
            points = points.T
            points = np.vstack((points, np.ones(points.shape[1])))

            points_camera_all = np.matmul(self.calib.T_camera_lidar, points)
            points_camera = np.array([])
            for i in range(points_camera_all.shape[1]):
                if points_camera_all[2, i] > 0:
                    points_camera = np.concatenate((points_camera, points_camera_all[:, i]))
            points_camera = np.reshape(points_camera, (-1, 4)).T
            pixel_camera = np.matmul(self.calib.P0, points_camera)

            image = copy.deepcopy(camera_data.load_data())

            max_z = int(max(pixel_camera[2, :]) / 3)
            for i in range(pixel_camera.shape[1]):
                z = pixel_camera[2, i]
                x = int(pixel_camera[0, i] / z)
                y = int(pixel_camera[1, i] / z)
                if x > 0 and x < image.shape[1] and y > 0 and y < image.shape[0]:
                    c = cv2.applyColorMap(np.array([int(pixel_camera[2, i] / max_z * 255)], dtype=np.uint8), cv2.COLORMAP_RAINBOW).squeeze().tolist()
                    cv2.circle(image, (x, y), 1, c, 1)

            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            ax.imshow(image)
            plt.draw()


import os.path as osp
import numpy as np


class Calib:
    """
    Class for loading and storing calibration matrices.
    """
    def __init__(self, calib_root):
        self.P0 = np.loadtxt(osp.join(calib_root, 'P_camera.txt'))
        self.T_applanix_lidar = np.loadtxt(osp.join(calib_root, 'T_applanix_lidar.txt'))
        self.T_camera_lidar = np.loadtxt(osp.join(calib_root, 'T_camera_lidar.txt'))
        self.T_radar_lidar = np.loadtxt(osp.join(calib_root, 'T_radar_lidar.txt'))

    def print_calibration(self):
        print('P0:')
        print(self.P0)
        print('T_applanix_lidar:')
        print(self.T_applanix_lidar)
        print('T_camera_lidar:')
        print(self.T_camera_lidar)
        print('T_radar_lidar:')
        print(self.T_radar_lidar)



import os.path as osp
import argparse
import struct
import matplotlib.pyplot as plt
import matplotlib
from datetime import datetime

# Periods during which daylight savings is in effect in Toronto
dst = [(1583632800, 1604196000), (1615687200, 1636250400), (1647136800, 1667700000), (1678586400, 1699236000)]

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--root', default='/mnt/data1/glen_shields/2020_12_01/', type=str,
                        help='location of root folder which contains applanix/smrmsg.out')
    args = parser.parse_args()
    with open(osp.join(args.root, 'applanix', 'smrmsg.out'), 'rb') as f:
        fc = f.read()
    with open(osp.join(args.root, 'applanix', 'ros_and_gps_time.csv'), 'r') as f:
        lines = f.readlines()
        start_time = float(lines[1].split(',')[1]) - 5.0
        end_time = float(lines[-1].split(',')[1]) + 5.0
    dt = datetime.fromtimestamp(start_time)
    g2 = dt.isoweekday() * 24 * 3600 + dt.hour * 3600 + dt.minute * 60 + dt.second + dt.microsecond * 1e-6
    start_week = round(start_time - g2)
    # get timezone offset:
    # Toronto time is GMT-4 or GMT-5 depending on time of year
    time_zone_offset = 5 * 3600
    for period in dst:
        if period[0] < start_time and start_time < period[1]:
            time_zone_offset = 4 * 3600

    start_gps = start_time + time_zone_offset - start_week
    end_gps = end_time + time_zone_offset - start_week
    
    t = []
    n = []
    e = []
    d = []
    vn = []
    ve = []
    vd = []
    r = []
    p = []
    h = []

    size = 10 * 8  # size of each line in bytes
    for i in range(len(fc) // size):
        data = struct.unpack("d" * 10, fc[i * size: (i + 1) * size])
        if data[0] < start_gps or data[0] > end_gps:
            continue
        t.append(data[0])
        n.append(data[1])
        e.append(data[2])
        d.append(data[3])
        vn.append(data[4])
        ve.append(data[5])
        vd.append(data[6])
        r.append(data[7])
        p.append(data[8])
        h.append(data[9])

    matplotlib.rcParams.update({'font.size': 16, 'xtick.labelsize': 16, 'ytick.labelsize': 16,
                                'axes.linewidth': 1.5, 'font.family': 'serif', 'pdf.fonttype': 42})

    plt.figure(figsize=(10, 5))
    plt.plot(t, n, label='North position RMS error (m)')
    plt.plot(t, e, label='East position RMS error (m)')
    plt.plot(t, d, label='Down position RMS error (m)')
    plt.grid(which='both', linestyle='--', alpha=0.5)
    plt.xlabel('time (seconds)', fontsize=16)
    plt.ylabel('Error (m)', fontsize=16)
    plt.legend(loc="upper left", prop={'size': 12})
    plt.xticks(rotation=45, ha="right")
    plt.savefig(osp.join(args.root, 'applanix', 'gps_position_error.pdf'), bbox_inches='tight', pad_inches=0.0)

    plt.figure(figsize=(10, 5))
    plt.plot(t, vn, label='North velocity RMS error (m/s)')
    plt.plot(t, ve, label='East velocity RMS error (m/s)')
    plt.plot(t, vd, label='Down velocity RMS error (m/s)')
    plt.grid(which='both', linestyle='--', alpha=0.5)
    plt.xlabel('time (seconds)', fontsize=16)
    plt.ylabel('Error (m / s)', fontsize=16)
    plt.legend(loc="upper left", prop={'size': 12})
    plt.xticks(rotation=45, ha="right")
    plt.savefig(osp.join(args.root, 'applanix', 'gps_velocity_error.pdf'), bbox_inches='tight', pad_inches=0.0)

    plt.figure(figsize=(10, 5))
    plt.plot(t, r, label='Roll RMS error (arc-min)')
    plt.plot(t, p, label='Pitch RMS error (arc-min)')
    plt.plot(t, h, label='Heading RMS error (arc-min)')
    plt.grid(which='both', linestyle='--', alpha=0.5)
    plt.xlabel('time (seconds)', fontsize=16)
    plt.ylabel('Error (arc-minutes)', fontsize=16)
    plt.legend(loc="upper left", prop={'size': 12})
    plt.xticks(rotation=45, ha="right")
    plt.savefig(osp.join(args.root, 'applanix', 'gps_orientation_error.pdf'), bbox_inches='tight', pad_inches=0.0)



################################################################################
#
# Copyright (c) 2017 University of Oxford
# Authors:
#  Dan Barnes (dbarnes@robots.ox.ac.uk)
#
# This work is licensed under the Creative Commons
# Attribution-NonCommercial-ShareAlike 4.0 International License.
# To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc-sa/4.0/ or send a letter to
# Creative Commons, PO Box 1866, Mountain View, CA 94042, USA.
#
###############################################################################

import numpy as np
import cv2
from pyboreas.utils.utils import get_time_from_filename

upgrade_time = 1632182400  # before: resolution = 0.0596, after: resolution = 0.04381


def load_radar(example_path):
    """Decode a single Oxford Radar RobotCar Dataset radar example
    Args:
        example_path (AnyStr): Oxford Radar RobotCar Dataset Example png
    Returns:
        timestamps (np.ndarray): Timestamp for each azimuth in int64 (UNIX time)
        azimuths (np.ndarray): Rotation for each polar radar azimuth (radians)
        valid (np.ndarray) Mask of whether azimuth data is an original sensor reading or interpolated from adjacent
            azimuths
        fft_data (np.ndarray): Radar power readings along each azimuth
    """
    # Hard coded configuration to simplify parsing code
    encoder_size = 5600
    resolution = 0.0596
    t = get_time_from_filename(example_path)
    if t > upgrade_time:
        resolution = 0.04381
    raw_example_data = cv2.imread(example_path, cv2.IMREAD_GRAYSCALE)
    timestamps = raw_example_data[:, :8].copy().view(np.int64)
    azimuths = (raw_example_data[:, 8:10].copy().view(np.uint16) / float(encoder_size) * 2 * np.pi).astype(np.float32)
    valid = raw_example_data[:, 10:11] == 255
    fft_data = raw_example_data[:, 11:].astype(np.float32)[:, :, np.newaxis] / 255.
    min_range = int(round(2.5 / resolution))
    fft_data[:, :min_range] = 0
    fft_data = np.squeeze(fft_data)
    return timestamps, azimuths, valid, fft_data, resolution


def radar_polar_to_cartesian(azimuths, fft_data, radar_resolution, cart_resolution, cart_pixel_width,
                             interpolate_crossover=True, fix_wobble=True):
    """Convert a polar radar scan to cartesian.
    Args:
        azimuths (np.ndarray): Rotation for each polar radar azimuth (radians)
        fft_data (np.ndarray): Polar radar power readings
        radar_resolution (float): Resolution of the polar radar data (metres per pixel)
        cart_resolution (float): Cartesian resolution (metres per pixel)
        cart_pixel_width (int): Width and height of the returned square cartesian output (pixels). Please see the Notes
            below for a full explanation of how this is used.
        interpolate_crossover (bool, optional): If true interpolates between the end and start  azimuth of the scan. In
            practice a scan before / after should be used but this prevents nan regions in the return cartesian form.

    Returns:
        np.ndarray: Cartesian radar power readings
    """
    if (cart_pixel_width % 2) == 0:
        cart_min_range = (cart_pixel_width / 2 - 0.5) * cart_resolution
    else:
        cart_min_range = cart_pixel_width // 2 * cart_resolution
    coords = np.linspace(-cart_min_range, cart_min_range, cart_pixel_width, dtype=np.float32)
    Y, X = np.meshgrid(coords, -1 * coords)
    sample_range = np.sqrt(Y * Y + X * X)
    sample_angle = np.arctan2(Y, X)
    sample_angle += (sample_angle < 0).astype(np.float32) * 2. * np.pi

    # Interpolate Radar Data Coordinates
    azimuth_step = (azimuths[-1] - azimuths[0]) / (azimuths.shape[0] - 1)
    sample_u = (sample_range - radar_resolution / 2) / radar_resolution
    sample_v = (sample_angle - azimuths[0]) / azimuth_step
    # This fixes the wobble in the old CIR204 data from Boreas
    EPS = 1e-14
    if fix_wobble and radar_resolution == 0.0596:
        azimuths = azimuths.reshape((1, 1, 400))  # 1 x 1 x 400
        sample_angle = np.expand_dims(sample_angle, axis=-1)  # H x W x 1
        diff = np.abs(azimuths - sample_angle)
        c3 = np.argmin(diff, axis=2)
        azimuths = azimuths.squeeze()
        c3 = c3.reshape(cart_pixel_width, cart_pixel_width)  # azimuth indices (closest)
        mindiff = sample_angle.squeeze() - azimuths[c3]
        sample_angle = sample_angle.squeeze()
        mindiff = mindiff.squeeze()

        subc3 = c3 * (c3 < 399)
        aplus = azimuths[subc3 + 1]
        a1 = azimuths[subc3]
        delta1 = mindiff * (mindiff > 0) * (c3 < 399) / (aplus - a1 + EPS)
        subc3 = c3 * (c3 > 0)
        a2 = azimuths[subc3]
        aminus = azimuths[1 + (c3 > 0) * (subc3 - 2)]
        delta2 = mindiff * (mindiff < 0) * (c3 > 0) / (a2 - aminus + EPS)
        sample_v = c3 + delta1 + delta2
        sample_v = sample_v.astype(np.float32)

    # We clip the sample points to the minimum sensor reading range so that we
    # do not have undefined results in the centre of the image. In practice
    # this region is simply undefined.
    sample_u[sample_u < 0] = 0

    if interpolate_crossover:
        fft_data = np.concatenate((fft_data[-1:], fft_data, fft_data[:1]), 0)
        sample_v = sample_v + 1

    polar_to_cart_warp = np.stack((sample_u, sample_v), -1)
    return cv2.remap(fft_data, polar_to_cart_warp, None, cv2.INTER_LINEAR)


def mean_intensity_mask(polar_data, multiplier=3.0):
    """Thresholds on multiplier*np.mean(azimuth_data) to create a polar mask of likely target points.
    Args:
        polar_data (np.ndarray): num_azimuths x num_range_bins polar data
        multiplier (float): multiple of mean that we treshold on
    Returns:
        np.ndarray: binary polar mask corresponding to likely target points
    """
    num_azimuths, range_bins = polar_data.shape
    mask = np.zeros((num_azimuths, range_bins))
    for i in range(num_azimuths):
        m = np.mean(polar_data[i, :])
        mask[i, :] = polar_data[i, :] > multiplier * m
    return mask



"""Tests for odometry benchmark"""
import unittest
import numpy as np

from pylgmath import Transformation, se3op
from utils.utils import get_inverse_tf
from utils.odometry import interpolate_poses, write_traj_file, read_traj_file


class OdometryTestCase(unittest.TestCase):
    """This class contains tests for the odometry benchmark script."""
    def test_interpolate(self):
        """Checks the pysteam interpolation against a known constant-velocity groundtruth."""
        # setup groundtruth
        delt = 0.05
        num_poses = 50

        # constant velocity
        v_x = -1.0
        omega_z = 0.01
        velocity_prior = np.array([[v_x, 0., 0., 0., 0., omega_z]]).T
        init_pose_vec = np.array([[0., 0., 0., 0., 0., 0.]]).T
        init_pose = Transformation(xi_ab=init_pose_vec)

        # create trajectory
        poses = [se3op.vec2tran(i*delt*velocity_prior)
                 @ init_pose.matrix() for i in range(num_poses)]
        times = [int(i*delt*1e9) for i in range(num_poses)]     # nanoseconds

        query_poses = interpolate_poses(poses[::2], times[::2], times[1:-1:2])
        gt_poses = poses[1:-1:2]
        for query_pose, gt_pose in zip(query_poses, gt_poses):
            delta = gt_pose@get_inverse_tf(query_pose)
            self.assertTrue(np.linalg.norm(se3op.tran2vec(delta)) < 1e-6)

    def test_read_write(self):
        """Checks read and write functions for trajectory."""
        # setup groundtruth
        delt = 0.05
        num_poses = 50

        # constant velocity
        v_x = -1.0
        omega_z = 0.01
        velocity_prior = np.array([[v_x, 0., 0., 0., 0., omega_z]]).T
        init_pose_vec = np.array([[0., 0., 0., 0., 0., 0.]]).T
        init_pose = Transformation(xi_ab=init_pose_vec)

        # create trajectory
        write_poses = [se3op.vec2tran(i*delt*velocity_prior)
                       @ init_pose.matrix() for i in range(num_poses)]
        write_times = [int(i*delt*1e9) for i in range(num_poses)]     # nanoseconds

        # write out trajectory to file, then read it back
        write_traj_file('test/test_traj.txt', write_poses, write_times)
        read_poses, read_times = read_traj_file('test/test_traj.txt')

        # compare write and read
        for wpose, wtime, rpose, rtime in zip(write_poses, write_times, read_poses, read_times):
            delta = wpose@get_inverse_tf(rpose)
            self.assertTrue(np.linalg.norm(se3op.tran2vec(delta)) < 1e-6)
            self.assertTrue(wtime == rtime)


if __name__ == '__main__':
    # run 'python -m unittest' from 'boreas-devkit/python' directory
    unittest.main()



import numpy as np
import multiprocessing
from multiprocessing import Pool
from pyboreas.utils.utils import se3ToSE3, is_sorted


class PointCloud:
    """
    Class for working with (lidar) pointclouds.
    """
    def __init__(self, points):
        # points (np.ndarray): (N, 6) [x, y, z, intensity, laser_number, time]
        self.points = points

    def transform(self, T, in_place=True):
        """Transforms points given a transform, T. x_out = np.matmul(T, x)
        Args:
            T (np.ndarray): 4x4 transformation matrix
            in_place (bool): if True, self.points is updated
        Returns:
            points (np.ndarray): The transformed points
        """
        assert (T.shape[0] == 4 and T.shape[1] == 4)
        if in_place:
            points = self.points
        else:
            points = np.copy(self.points)
        p = np.hstack((points[:, :3], np.ones((points.shape[0], 1))))
        points[:, :3] = np.matmul(p, T.transpose())[:, :3]
        return points

    def remove_motion(self, body_rate, tref=None, in_place=True):
        """Removes motion distortion from a pointcloud
        Args:
            body_rate (np.ndarray): (6, 1) [vx, vy, vz, wx, wy, wz] in sensor frame
            tref (float): reference time to transform the points towards
            in_place (bool): if True, self.points is updated
        Returns:
            points (np.ndarray): points with motion distortion removed
        """
        assert(body_rate.shape[0] == 6 and body_rate.shape[1] == 1)
        tmin = np.min(self.points[:, 5])
        tmax = np.max(self.points[:, 5])
        if tref is None:
            tref = (tmin + tmax) / 2
        # Precompute finite number of transforms for speed
        bins = 21
        delta = (tmax - tmin) / float(bins - 1)
        T_undistorts = []
        tbins = []
        for i in range(bins):
            t = tmin + i * delta
            tbins.append(t)
            T_undistorts.append(se3ToSE3((t - tref) * body_rate))
        tbins[-1] += 1e-6

        if in_place:
            points = self.points
        else:
            points = np.copy(self.points)

        global _process_motion

        def _process_motion(i):
            index = int((points[i, 5] - tmin) / delta)
            pbar = np.vstack((points[i, :3].reshape(3, 1), 1))
            pbar = np.matmul(T_undistorts[index], pbar)
            points[i, :3] = pbar[:3, 0]

        # This function is ~10x faster if the pointcloud is sorted by timestamp (default)
        sections = []
        if is_sorted(points[:, 5]):
            for i in range(len(tbins) - 1):
                locs = np.where((points[:, 5] >= tbins[i]) & (points[:, 5] < tbins[i+1]))
                p = points[locs]
                p = np.hstack((p[:, :3], np.ones((p.shape[0], 1))))
                p = np.matmul(p, T_undistorts[i].transpose())
                p = np.hstack((p[:, :3], points[locs][:, 3:]))
                points[locs] = p
        else:
            p = Pool(multiprocessing.cpu_count())
            p.map(_process_motion, list(range(points.shape[0])))

        if in_place:
            self.points = points

        return points

    def passthrough(self, bounds=[], in_place=True):
        """Removes points outside the specified bounds
        Args:
            bounds (list): [xmin, xmax, ymin, ymax, zmin, zmax]
            in_place (bool): if True, self.points is updated
        Returns:
            points (np.ndarray): the remaining points after the filter is applied
        """
        if len(bounds) < 6:
            print('Warning: len(bounds) = {} < 6 is incorrect!'.format(len(bounds)))
            return self.points
        p = self.points[np.where((self.points[:, 0] >= bounds[0]) &
                                 (self.points[:, 0] <= bounds[1]) &
                                 (self.points[:, 1] >= bounds[2]) &
                                 (self.points[:, 1] <= bounds[3]) &
                                 (self.points[:, 2] >= bounds[4]) &
                                 (self.points[:, 2] <= bounds[5]))]
        if in_place:
            self.points = p
        return p

    # Assumes pointcloud has already been transformed into the camera frame
    # color options: depth, intensity
    # returns pixel locations for lidar point projections onto an image plane
    def project_onto_image(self, P, width=2448, height=2048, color='depth'):
        """Projects 3D points onto a 2D image plane
        Args:
            P (np.ndarray): [fx 0 cx 0; 0 fy cy 0; 0 0 1 0; 0 0 0 1] cam projection
            width (int): width of image
            height (int): height of image
            color (str): 'depth' or 'intensity' to pick colors output
        Return:
            uv (np.ndarray): (N, 2) projected u-v pixel locations in an image
            colors (np.ndarray): (N,) a color value for each pixel location in uv.
        """
        uv = []
        colors = []
        x = np.hstack((self.points[:, :3], np.ones((self.points.shape[0], 1))))
        x /= x[:, 2:3]
        x[:, 3] = 1
        x = np.matmul(x, P.transpose())
        mask = np.where((x[:, 0] >= 0) &
                        (x[:, 0] <= width - 1) &
                        (x[:, 1] >= 0) &
                        (x[:, 1] <= height - 1))
        x = x[mask]
        if color == 'depth':
            colors = self.points[mask][:, 2]
        elif color == 'intensity':
            colors = self.points[mask][:, 3]
        else:
            print('Warning: {} is not a valid color'.format(color))
            colors = self.points[mask][:, 2]
        return x[:, :2], colors

# TODO: remove_ground(self, bool: in_place)
# TODO: voxelize(self)


import os
from time import time
from itertools import accumulate
import numpy as np
import matplotlib.pyplot as plt
from pyboreas.utils.utils import get_inverse_tf, rotation_error, translation_error, enforce_orthog
from pylgmath import Transformation
from pysteam.trajectory import Time, TrajectoryInterface
from pysteam.state import TransformStateVar, VectorSpaceStateVar
from pysteam.problem import OptimizationProblem
from pysteam.solver import GaussNewtonSolver
from pysteam.evaluator import TransformStateEvaluator


class TrajStateVar:
    """This class defines a trajectory state variable for steam."""
    def __init__(
          self,
          time: Time,
          pose: TransformStateVar,
          velocity: VectorSpaceStateVar,
    ) -> None:
        self.time: Time = time
        self.pose: TransformStateVar = pose
        self.velocity: VectorSpaceStateVar = velocity


def interpolate_poses(poses, times, query_times, init_finite_diff=False):
    """Runs a steam optimization with locked poses and outputs poses queried at query_times
    Args:
        poses (List[np.ndarray]): list of 4x4 poses (T_v_i vehicle and inertial frames)
        times (List[int]): list of times for poses (float for seconds or int for nanoseconds)
        query_times (List[float or int]): list of query times (int for nanoseconds)
        init_finite_diff (bool): if set to true, initialize velocities using finite difference of poses
    Returns:
        (List[np.ndarray]): list of 4x4 poses (T_v_i vehicle and inertial frames) at query_times
    """
    # smoothing factor diagonal
    Qc_inv = np.diag(1 / np.array([1.0, 0.001, 0.001, 0.001, 0.001, 1.0]))

    # steam state variables
    if init_finite_diff:
        states = []
        for i in range(len(poses)):
            if i == 0:
                dt = (times[1] - times[0])*1e-9
                dT = Transformation(T_ba=enforce_orthog(poses[1])) @ Transformation(T_ba=enforce_orthog(poses[0]))
            else:
                dt = (times[i] - times[i-1])*1e-9
                dT = Transformation(T_ba=enforce_orthog(poses[i])) @ Transformation(T_ba=enforce_orthog(poses[i-1]))
            velocity = dT.vec() / dt
            states += [TrajStateVar(
                            Time(nsecs=times[i]),
                            TransformStateVar(Transformation(T_ba=enforce_orthog(poses[i])), copy=True),
                            VectorSpaceStateVar(velocity, copy=True))]
    else:
        init_velocity = np.zeros((6, 1))
        states = [
            TrajStateVar(
                Time(nsecs=times[i]),
                TransformStateVar(Transformation(T_ba=enforce_orthog(poses[i])), copy=True),
                VectorSpaceStateVar(init_velocity, copy=True),
            ) for i in range(len(poses))
        ]

    # setup trajectory
    traj = TrajectoryInterface(Qc_inv=Qc_inv, allow_extrapolation=True)
    for state in states:
        traj.add_knot(time=state.time, T_k0=TransformStateEvaluator(state.pose), velocity=state.velocity)
        state.pose.set_lock(True)   # lock all pose variables

    # construct the optimization problem
    opt_prob = OptimizationProblem()
    opt_prob.add_cost_term(*traj.get_prior_cost_terms())
    opt_prob.add_state_var(*[j for i in states for j in (i.pose, i.velocity)])

    # construct the solver
    optimizer = GaussNewtonSolver(opt_prob, verbose=False)

    # solve the problem (states are automatically updated)
    optimizer.optimize()

    query_poses = []
    for time in query_times:
        interp_eval = traj.get_interp_pose_eval(Time(nsecs=time))
        query_poses += [interp_eval.evaluate().matrix()]

    return query_poses


def trajectory_distances(poses):
    """Calculates path length along the trajectory.
    Args:
        poses (List[np.ndarray]): list of 4x4 poses (T_k_i, 'i' is a fixed reference frame)
    Returns:
        List[float]: distance along the trajectory, increasing as a function of time / list index
    """
    dist = [0]
    for i in range(1, len(poses)):
        P1 = get_inverse_tf(poses[i - 1])
        P2 = get_inverse_tf(poses[i])
        dx = P1[0, 3] - P2[0, 3]
        dy = P1[1, 3] - P2[1, 3]
        dz = P1[2, 3] - P2[2, 3]
        dist.append(dist[i-1] + np.sqrt(dx**2 + dy**2 + dz**2))
    return dist


def last_frame_from_segment_length(dist, first_frame, length):
    """Retrieves the index of the last frame for our current analysis.
        last_frame should be 'dist' meters away from first_frame in terms of distance traveled along the trajectory.
    Args:
        dist (List[float]): distance along the trajectory, increasing as a function of time / list index
        first_frame (int): index of the starting frame for this sequence
        length (float): length of the current segment being evaluated
    Returns:
        last_frame (int): index of the last frame in this segment
    """
    for i in range(first_frame, len(dist)):
        if dist[i] > dist[first_frame] + length:
            return i
    return -1


def calc_sequence_errors(poses_gt, poses_pred, step_size=4):
    """Calculate the translation and rotation error for each subsequence across several different lengths.
    Args:
        T_gt (List[np.ndarray]): each entry in list is 4x4 transformation matrix, ground truth transforms
        T_pred (List[np.ndarray]): each entry in list is 4x4 transformation matrix, predicted transforms
    Returns:
        err (List[Tuple]): each entry in list is [first_frame, r_err, t_err, length, speed]
        lengths (List[int]): list of lengths that odometry is evaluated at
    """
    lengths = [100, 200, 300, 400, 500, 600, 700, 800]
    err = []
    # Pre-compute distances from ground truth as reference
    dist = trajectory_distances(poses_gt)

    for first_frame in range(0, len(poses_gt), step_size):
        for length in lengths:
            last_frame = last_frame_from_segment_length(dist, first_frame, length)
            if last_frame == -1:
                continue
            # Compute rotational and translation errors
            pose_delta_gt = np.matmul(poses_gt[last_frame], get_inverse_tf(poses_gt[first_frame]))
            pose_delta_res = np.matmul(poses_pred[last_frame], get_inverse_tf(poses_pred[first_frame]))
            pose_error = np.matmul(pose_delta_gt, get_inverse_tf(pose_delta_res))
            r_err = rotation_error(pose_error)
            t_err = translation_error(pose_error)
            # Approx speed
            num_frames = float(last_frame - first_frame + 1)
            speed = float(length) / (0.1 * num_frames)
            err.append([first_frame, r_err/float(length), t_err/float(length), length, speed])
    return err, lengths


def get_stats(err, lengths):
    """Computes the average translation and rotation within a sequence (across subsequences of diff lengths).
    Args:
        err (List[Tuple]): each entry in list is [first_frame, r_err, t_err, length, speed]
        lengths (List[int]): list of lengths that odometry is evaluated at
    Returns:
        average translation (%) and rotation (deg/m) errors
    """
    t_err = 0
    r_err = 0
    len2id = {x: i for i, x in enumerate(lengths)}
    t_err_len = [0.0]*len(len2id)
    r_err_len = [0.0]*len(len2id)
    len_count = [0]*len(len2id)
    for e in err:
        t_err += e[2]
        r_err += e[1]
        j = len2id[e[3]]
        t_err_len[j] += e[2]
        r_err_len[j] += e[1]
        len_count[j] += 1
    t_err /= float(len(err))
    r_err /= float(len(err))
    return t_err * 100, r_err * 180 / np.pi, [a/float(b) * 100 for a, b in zip(t_err_len, len_count)], \
        [a/float(b) * 180 / np.pi for a, b in zip(r_err_len, len_count)]


def plot_stats(seq, root, T_odom, T_gt, lengths, t_err, r_err):
    """Outputs plots of calculated statistics to specified directory.
    Args:
        seq (List[string]): list of sequence file names
        root (string): directory path for plot outputs
        T_odom (List[np.ndarray]): list of 4x4 estimated poses T_vk_i (vehicle frame at time k and fixed frame i)
        T_gt (List[np.ndarray]): List of 4x4 groundtruth poses T_vk_i (vehicle frame at time k and fixed frame i)
        lengths (List[int]): list of lengths that odometry is evaluated at
        t_err (List[float]): list of average translation error corresponding to lengths
        r_err (List[float]): list of average rotation error corresponding to lengths
    """
    path_odom = get_path_from_Tvi_list(T_odom)
    path_gt = get_path_from_Tvi_list(T_gt)

    # plot of path
    plt.figure(figsize=(6, 6))
    plt.plot(path_odom[:, 0], path_odom[:, 1], 'b', linewidth=0.5, label='Estimate')
    plt.plot(path_gt[:, 0], path_gt[:, 1], 'r', linewidth=0.5, label='Groundtruth')
    plt.plot(path_gt[0, 0], path_gt[0, 1], 'ks', markerfacecolor='none', label='Sequence Start')
    plt.xlabel('x [m]')
    plt.ylabel('y [m]')
    plt.axis('equal')
    plt.legend(loc="upper right")
    plt.savefig(os.path.join(root, seq[:-4] + '_path.pdf'), bbox_inches='tight')
    plt.close()

    # plot of translation error along path length
    plt.figure(figsize=(6, 3))
    plt.plot(lengths, t_err, 'bs', markerfacecolor='none')
    plt.plot(lengths, t_err, 'b')
    plt.xlabel('Path Length [m]')
    plt.ylabel('Translation Error [%]')
    plt.savefig(os.path.join(root, seq[:-4] + '_tl.pdf'), bbox_inches='tight')
    plt.close()

    # plot of rotation error along path length
    plt.figure(figsize=(6, 3))
    plt.plot(lengths, r_err, 'bs', markerfacecolor='none')
    plt.plot(lengths, r_err, 'b')
    plt.xlabel('Path Length [m]')
    plt.ylabel('Rotation Error [deg/m]')
    plt.savefig(os.path.join(root, seq[:-4] + '_rl.pdf'), bbox_inches='tight')
    plt.close()


def get_path_from_Tvi_list(Tvi_list):
    """Gets 3D path (xyz) from list of poses T_vk_i (transform between vehicle frame at time k and fixed frame i).
    Args:
        Tvi_list (List[np.ndarray]): K length list of 4x4 poses T_vk_i
    Returns:
        path (np.ndarray): K x 3 numpy array of xyz coordinates
    """
    path = np.zeros((len(Tvi_list), 3), dtype=np.float32)
    for j, Tvi in enumerate(Tvi_list):
        path[j] = (-Tvi[:3, :3].T @ Tvi[:3, 3:4]).squeeze()
    return path


def compute_kitti_metrics(T_gt, T_pred, times_gt, times_pred, seq_lens_gt, seq_lens_pred, seq, root, step_size=10):
    """Computes the translational (%) and rotational drift (deg/m) in the KITTI style.
        KITTI rotation and translation metrics are computed for each sequence individually and then
        averaged across the sequences.
    Args:
        T_gt (List[np.ndarray]): List of 4x4 homogeneous transforms (fixed reference frame to frame t)
        T_pred (List[np.ndarray]): List of 4x4 homogeneous transforms (fixed reference frame to frame t)
        times_gt (List[int]): List of times (nanoseconds) corresponding to T_gt
        times_pred (List[int]): List of times (nanoseconds) corresponding to T_pred
        seq_lens_gt (List[int]): List of sequence lengths corresponding to T_gt
        seq_lens_pred (List[int]): List of sequence lengths corresponding to T_pred
        seq (List[string]): List of sequence file names
        root (string): path to output directory for plots
        step_size (int): step size applied for classifying distances travelled
    Returns:
        t_err: Average KITTI Translation ERROR (%)
        r_err: Average KITTI Rotation Error (deg / m)
    """
    # get start and end indices of each sequence
    indices_gt = [0]
    indices_gt.extend(list(accumulate(seq_lens_gt)))
    indices_pred = [0]
    indices_pred.extend(list(accumulate(seq_lens_pred)))

    # loop for each sequence
    err_list = []
    for i in range(len(seq_lens_pred)):
        ts = time()  # start time
        print('processing sequence', seq[i], '...')
        # get poses and times of current sequence
        T_gt_seq = T_gt[indices_gt[i]:indices_gt[i+1]]
        T_pred_seq = T_pred[indices_pred[i]:indices_pred[i+1]]
        times_gt_seq = times_gt[indices_gt[i]:indices_gt[i+1]]
        times_pred_seq = times_pred[indices_pred[i]:indices_pred[i+1]]

        # query predicted trajectory at groundtruth times
        T_query = interpolate_poses(T_pred_seq, times_pred_seq, times_gt_seq)

        err, path_lengths = calc_sequence_errors(T_gt_seq, T_query, step_size)
        t_err, r_err, t_err_len, r_err_len = get_stats(err, path_lengths)
        err_list.append([t_err, r_err])

        print(seq[i], 'took', str(time() - ts), ' seconds')

        plot_stats(seq[i], root, T_query, T_gt_seq, path_lengths, t_err_len, r_err_len)
    err_list = np.asarray(err_list)
    avg = np.mean(err_list, axis=0)
    t_err = avg[0]
    r_err = avg[1]

    return t_err, r_err


def get_sequences(path, file_ext=''):
    """Retrieves a list of all the sequences in the dataset with the given prefix.
    Args:
        path (string): directory path to where the files are
        file_ext (string): string identifier to look for (e.g., '.txt')
    Returns:
        sequences (List[string]): list of sequence file names
    """
    sequences = [f for f in os.listdir(path) if file_ext in f]
    sequences.sort()
    return sequences


def get_sequence_poses(path, seq):
    """Retrieves a list of the poses corresponding to the given sequences in the given file path.
    Args:
        path (string): directory path to where the files are
        seq (List[string]): list of sequence file names
    Returns:
        all_poses (List[np.ndarray]): list of 4x4 poses from all sequence files
        all_times (List[int]): list of times in nanoseconds from all sequence files
        seq_lens (List[int]): list of sequence lengths
    """

    # loop for each sequence
    all_poses = []
    all_times = []
    seq_lens = []
    for filename in seq:
        # parse file for list of poses and times
        poses, times = read_traj_file(os.path.join(path, filename))
        seq_lens.append(len(times))
        all_poses.extend(poses)
        all_times.extend(times)

    return all_poses, all_times, seq_lens


def write_traj_file(path, poses, times):
    """Writes trajectory into a space-separated txt file
    Args:
        path (string): file path including file name
        poses (List[np.ndarray]): list of 4x4 poses (T_v_i vehicle and inertial frames)
        times (List[int]): list of times for poses (int for nanoseconds)
    """
    with open(path, "w") as file:
        # Writing each time (nanoseconds) and pose to file
        for time, pose in zip(times, poses):
            line = [time]
            line.extend(pose.reshape(16)[:12].tolist())
            file.write(' '.join(str(num) for num in line))
            file.write('\n')


def read_traj_file(path):
    """Writes trajectory into a space-separated txt file
    Args:
        path (string): file path including file name
    Returns:
        (List[np.ndarray]): list of 4x4 poses
        (List[int]): list of times in nanoseconds
    """
    with open(path, "r") as file:
        # read each time and pose to lists
        poses = []
        times = []

        for line in file:
            line_split = line.strip().split()
            values = [float(v) for v in line_split[1:]]
            pose = np.zeros((4, 4))
            pose[0, 0:4] = values[0:4]
            pose[1, 0:4] = values[4:8]
            pose[2, 0:4] = values[8:12]
            pose[3, 3] = 1.0
            poses.append(pose)
            times.append(int(line_split[0]))

    return poses, times



import os
import os.path as osp

from pyboreas.data.calib import Calib
from pyboreas.data.sensors import Camera, Lidar, Radar
from pyboreas.utils.utils import get_closest_index


class Sequence:
    """
    Class for working with an individual Boreas dataset sequence
    """
    def __init__(self, boreas_root, seqSpec, verbose=False):
        """init
        Args:
            boreas_root (str): path to root folder ex: /path/to/data/boreas/
            seqSpec (list): defines sequence ID, start_time, and end_time
            verbose (bool): whether or not to print info during initialization
        """
        self.ID = seqSpec[0]
        self.verbose = verbose
        if verbose:
            print('SEQ: {}'.format(self.ID))
        if len(seqSpec) > 2:
            assert seqSpec[2] > seqSpec[1], 'Sequence timestamps must go forward in time'
            self.start_ts = str(seqSpec[1])
            self.end_ts = str(seqSpec[2])
            if verbose:
                print('START: {} END: {}'.format(self.start_ts, self.end_ts))
        else:
            self.start_ts = '0'  # dummy start and end if not specified
            self.end_ts = '9' * 21
        self.seq_root = osp.join(boreas_root, self.ID)
        self.applanix_root = osp.join(self.seq_root, 'applanix')
        self.calib_root = osp.join(self.seq_root, 'calib')
        self.camera_root = osp.join(self.seq_root, 'camera')
        self.lidar_root = osp.join(self.seq_root, 'lidar')
        self.radar_root = osp.join(self.seq_root, 'radar')

        self._check_dataroot_valid()  # Check if folder structure correct

        self.calib = Calib(self.calib_root)
        # Creates list of frame objects for cam, lidar, radar, and inits poses
        self.get_all_frames()

        self._check_download()  # if verbose, prints warning when sensor data missing

        if verbose:
            print('camera frames: {}'.format(len(self.camera_frames)))
            print('lidar frames: {}'.format(len(self.lidar_frames)))
            print('radar frames: {}'.format(len(self.radar_frames)))
            print('-------------------------------')

    def get_camera(self, idx):
        self.camera_frames[idx].load_data()
        return self.camera_frames[idx]

    @property
    def camera(self):
        for camera_frame in self.camera_frames:
            camera_frame.load_data()
            yield camera_frame

    def get_camera_iter(self):
        """Retrieves an iterator on camera frames"""
        return iter(self.camera)

    def get_lidar(self, idx):
        self.lidar_frames[idx].load_data()
        return self.lidar_frames[idx]

    @property
    def lidar(self):
        for lidar_frame in self.lidar_frames:
            lidar_frame.load_data()
            yield lidar_frame

    def get_lidar_iter(self):
        """Retrieves an iterator on lidar frames"""
        return iter(self.lidar)

    def get_radar(self, idx):
        self.radar_frames[idx].load_data()
        return self.radar_frames[idx]

    @property
    def radar(self):
        for radar_frame in self.radar_frames:
            radar_frame.load_data()
            yield radar_frame

    def get_radar_iter(self):
        """Retrieves an iterator on radar frames"""
        return iter(self.radar)

    def _check_dataroot_valid(self):
        """Checks if the sequence folder structure is valid"""
        if not osp.isdir(self.applanix_root):
            raise ValueError("ERROR: applanix dir missing from dataroot")
        if not osp.isdir(self.calib_root):
            raise ValueError("ERROR: calib dir missing from dataroot")
        if not osp.isdir(self.camera_root):
            os.mkdir(self.camera_root)
        if not osp.isdir(self.lidar_root):
            os.mkdir(self.lidar_root)
        if not osp.isdir(self.radar_root):
            os.mkdir(self.radar_root)

    def _check_download(self):
        """Checks if all sensor data has been downloaded, prints a warning otherwise"""
        if len(os.listdir(self.camera_root)) < len(self.camera_frames):
            print('WARNING: camera images are not all downloaded')
        if len(os.listdir(self.lidar_root)) < len(self.lidar_frames):
            print('WARNING: lidar frames are not all downloaded')
        if len(os.listdir(self.radar_root)) < len(self.radar_frames):
            print('WARNING: radar scans are not all downloaded')

    def _get_frames(self, posefile, root, ext, SensorType):
        """Initializes sensor frame objects with their ground truth pose information
        Args:
            posefile (str): path to ../sensor_poses.csv
            root (str): path to the root of the sensor folder ../sensor/
            ext (str): file extension specific to this sensor type
            SensorType (cls): sensor class specific to this sensor type
        Returns:
            frames (list): list of sensor frame objects
        """
        frames = []
        if osp.exists(posefile) and osp.isdir(root):
            with open(posefile, 'r') as f:
                f.readline()  # header
                for line in f:
                    data = line.split(',')
                    ts = data[0]
                    if self.start_ts <= ts and ts <= self.end_ts:
                        frame = SensorType(osp.join(root, ts + ext))
                        frame.init_pose(data)
                        frames.append(frame)
        return frames

    def get_all_frames(self):
        """Convenience method for retrieving sensor frames of all types"""
        cfile = osp.join(self.applanix_root, 'camera_poses.csv')
        lfile = osp.join(self.applanix_root, 'lidar_poses.csv')
        rfile = osp.join(self.applanix_root, 'radar_poses.csv')
        self.camera_frames = self._get_frames(cfile, self.camera_root, '.png', Camera)
        self.lidar_frames = self._get_frames(lfile, self.lidar_root, '.bin', Lidar)
        self.radar_frames = self._get_frames(rfile, self.radar_root, '.png', Radar)

    def reset_frames(self):
        """Resets all frames, removes downloaded data"""
        self.get_all_frames()

    def synchronize_frames(self, ref='camera'):
        """Simulates having synchronous measurements
        Note: measurements still won't be at the exact same timestamp and will have different poses
        However, for a given reference index, the other measurements will be as close to the reference
        in time as they can be.
        Args:
            ref (str): [camera, lidar, or radar] this determines which sensor's frames will be used as the
                reference for synchronization. This sensor's list of frames will not be modified. However,
                the other two list of sensor frames will be modified so that each index will approximately
                align with the reference in time.
        """
        cstamps = [frame.timestamp for frame in self.camera_frames]
        lstamps = [frame.timestamp for frame in self.lidar_frames]
        rstamps = [frame.timestamp for frame in self.radar_frames]

        def get_closest(query_time, target_times, targets):
            closest = get_closest_index(query_time, target_times)
            assert(abs(query_time - target_times[closest]) < 1.0), 'query: {}'.format(query_time)
            return targets[closest]

        if ref == 'camera':
            self.lidar_frames = [get_closest(cstamp, lstamps, self.lidar_frames) for cstamp in cstamps]
            self.radar_frames = [get_closest(cstamp, rstamps, self.radar_frames) for cstamp in cstamps]
        elif ref == 'lidar':
            self.camera_frames = [get_closest(lstamp, cstamps, self.camera_frames) for lstamp in lstamps]
            self.radar_frames = [get_closest(lstamp, rstamps, self.radar_frames) for lstamp in lstamps]
        elif ref == 'radar':
            self.camera_frames = [get_closest(rstamp, cstamps, self.camera_frames) for rstamp in rstamps]
            self.lidar_frames = [get_closest(rstamp, lstamps, self.lidar_frames) for rstamp in rstamps]


import numpy as np


class BoundingBox2D:
    def __init__(self, position, rotation, extent, label=None):
        """Checks dimensional consistency of inputs and constructs points array

        Args:
            position: (x,y,z) position of bbox centroid
            rotation: rotation matrix for bbox orientation
            extent: (width, length, height) of the bbox
            label: optional string for bbox label
        """
        assert position.shape == (3, 1)
        assert rotation.shape == (3, 3)
        assert extent.shape == (3, 1)
        self.pos = position
        self.rot = rotation
        self.extent = extent
        self.label = label

        # Construct array to extract points from extent
        dims_multiplier = np.array([[1, 1, 1], [-1, 1, 1], [-1, -1, 1], [1, -1, 1],
                                    [1, 1, -1], [-1, 1, -1], [-1, -1, -1], [1, -1, -1]])

        # TODO: need orientation here
        points = []
        for i in range(dims_multiplier.shape[0]):
            points.append(position + extent * dims_multiplier[i].reshape(3, 1) / 2)
        self.points = np.concatenate(points, axis=-1)  # Top 4 points ccw, then bottom 4 points, ccw. Dims=(3, #pts)

    def render_bbox_2d(self, ax, color="r"):
        """Render the bbox into a top-down 2d view

        Args:
            ax: the axis to render the bbox onto
        """
        prev_pt = self.points[:, 3]
        for i in range(4):  # Just draw top 4 points of bbox
            ax.plot([prev_pt[0], self.points[0, i]], [prev_pt[1], self.points[1, i]], color=color)
            prev_pt = self.points[:, i]

    def get_points_perspective(self, view: np.ndarray, normalize: bool) -> np.ndarray:
        """
        This is a helper class that maps 3d points to a 2d plane. It can be used to implement both perspective and
        orthographic projections. It first applies the dot product between the points and the view. By convention,
        the view should be such that the data is projected onto the first 2 axis. It then optionally applies a
        normalization along the third dimension.

        For a perspective projection the view should be a 3x3 camera matrix, and normalize=True
        For an orthographic projection with translation the view is a 3x4 matrix and normalize=False
        For an orthographic projection without translation the view is a 3x3 matrix (optionally 3x4 with last columns
         all zeros) and normalize=False

        :param points: <np.float32: 3, n> Matrix of points, where each point (x, y, z) is along each column.
        :param view: <np.float32: n, n>. Defines an arbitrary projection (n <= 4).
            The projection should be such that the corners are projected onto the first 2 axis.
        :param normalize: Whether to normalize the remaining coordinate (along the third axis).
        :return: <np.float32: 3, n>. Mapped point. If normalize=False, the third coordinate is the height.
        """

        assert view.shape[0] <= 4
        assert view.shape[1] <= 4

        viewpad = np.eye(4)
        viewpad[:view.shape[0], :view.shape[1]] = view

        nbr_points = self.points.shape[1]

        # Do operation in homogenous coordinates.
        points = np.concatenate((self.points, np.ones((1, nbr_points))))
        points = np.dot(viewpad, points)
        points = points[:3, :]

        if normalize:
            points = points / points[2:3, :].repeat(3, 0).reshape(3, nbr_points)

        return points

    def get_raw_data(self):
        return self.pos, self.rot, self.extent



import os.path as osp
from pathlib import Path
import numpy as np
import cv2


def load_lidar(path):
    """Loads a pointcloud (np.ndarray) (N, 6) from path [x, y, z, intensity, laser_number, time]"""
    points = np.fromfile(path, dtype=np.float32).reshape((-1, 6)).astype(np.float64)
    t = get_time_from_filename(path)
    points[:, 5] += t
    return points


def roll(r):
    return np.array([[1, 0, 0], [0, np.cos(r), np.sin(r)], [0, -np.sin(r), np.cos(r)]], dtype=np.float64)


def pitch(p):
    return np.array([[np.cos(p), 0, -np.sin(p)], [0, 1, 0], [np.sin(p), 0, np.cos(p)]], dtype=np.float64)


def yaw(y):
    return np.array([[np.cos(y), np.sin(y), 0], [-np.sin(y), np.cos(y), 0], [0, 0, 1]], dtype=np.float64)


def yawPitchRollToRot(y, p, r):
    Y = yaw(y)
    P = pitch(p)
    R = roll(r)
    C = np.matmul(P, Y)
    return np.matmul(R, C)


def rotToYawPitchRoll(C):
    i = 2
    j = 1
    k = 0
    c_y = np.sqrt(C[i, i]**2 + C[j, i]**2)
    if c_y > 1e-14:
        r = np.arctan2(C[j, i], C[i, i])
        p = np.arctan2(-C[k, i], c_y)
        y = np.arctan2(C[k, j], C[k, k])
    else:
        r = 0
        p = np.arctan2(-C[k, i], c_y)
        y = np.arctan2(-C[j, k], C[j, j])
    return y, p, r


def get_transform(gt):
    """Retrieve 4x4 homogeneous transform for a given parsed line of the ground truth pose csv
    Args:
        gt (List[float]): parsed line from ground truth csv file
    Returns:
        np.ndarray: 4x4 transformation matrix (pose of sensor)
    """
    T = np.identity(4, dtype=np.float64)
    C_enu_sensor = yawPitchRollToRot(gt[9], gt[8], gt[7])
    T[0, 3] = gt[1]
    T[1, 3] = gt[2]
    T[2, 3] = gt[3]
    T[0:3, 0:3] = C_enu_sensor
    return T


def get_transform2(R, t):
    """Returns a 4x4 homogeneous 3D transform
    Args:
        R (np.ndarray): (3,3) rotation matrix
        t (np.ndarray): (3,1) translation vector
    Returns:
        np.ndarray: 4x4 transformation matrix
    """
    T = np.identity(4, dtype=R.dtype)
    T[0:3, 0:3] = R
    T[0:3, 3] = t.squeeze()
    return T


def get_transform3(x, y, theta, dtype=np.float32):
    """Returns a 4x4 homogeneous 3D transform for a given 2D (x, y, theta).
    Args:
        x (float): x-translation
        y (float): y-translation
        theta (float): rotation
    Returns:
        np.ndarray: 4x4 transformation matrix
    """
    T = np.identity(4, dtype=dtype)
    T[0:2, 0:2] = np.array([[np.cos(theta), np.sin(theta)], [-np.sin(theta), np.cos(theta)]])
    T[0, 3] = x
    T[1, 3] = y
    return T


def quaternionToRot(qin):
    """Converts a quaternion to a rotation  matrix
    Args:
        qin (np.ndarray) (4,) [qx, qy, qz, qw] quaternion
    Returns:
        C (np.ndarray) (3,3) rotation matrix
    """
    q = qin.copy().reshape(4, 1)
    if np.matmul(q.transpose(), q) < 1e-14:
        return np.identity(3)
    xi = q[:3].reshape(3, 1)
    eta = q[3, 0]
    C = (eta**2 - np.matmul(xi.transpose(), xi)) * np.identity(3) + \
        2 * np.matmul(xi, xi.transpose()) - 2 * eta * carrot(xi)
    return C


def rotToQuaternion(C):
    """Converts a rotation matrix to a quaternion
    Note that the space of unit-length quaternions is a double-cover of SO(3)
    which means that, C maps to +/- q, so q --> C --> +/- q
    Args:
        C (np.ndarray) (3,3) rotation matrix
    Returns:
        q (np.ndarray) (4,1) [qx, qy, qz, qw] quaternion
    """
    eta = 0.5 * np.sqrt((1 + np.trace(C)))
    if np.abs(eta) < 1e-14:
        eta = 0
        xi = np.sqrt(np.diag(0.5 * (C + np.identity(3))))
        q = np.array([xi[0], xi[1], xi[2], eta]).reshape(4, 1)
    else:
        phi = wrapto2pi(2 * np.arccos(eta))
        eta = np.cos(phi / 2)
        xi_cross = (C.T - C) / (4 * eta)
        q = np.array([xi_cross[2, 1], xi_cross[0, 2], xi_cross[1, 0], eta]).reshape(4, 1)
    return q


def get_inverse_tf(T):
    """Returns the inverse of a given 4x4 homogeneous transform.
    Args:
        T (np.ndarray): 4x4 transformation matrix
    Returns:
        np.ndarray: inv(T)
    """
    T2 = np.identity(4, dtype=T.dtype)
    R = T[:3, :3]
    t = T[:3, 3:]
    T2[:3, :3] = R.transpose()
    T2[:3, 3:] = np.matmul(-1 * R.transpose(), t)
    return T2


def enforce_orthog(T, dim=3):
    """Enforces orthogonality of a 3x3 rotation matrix within a 4x4 homogeneous transformation matrix.
    Args:
        T (np.ndarray): 4x4 transformation matrix
        dim (int): dimensionality of the transform 2==2D, 3==3D
    Returns:
        np.ndarray: 4x4 transformation matrix with orthogonality conditions on the rotation matrix enforced.
    """
    if dim == 2:
        if abs(np.linalg.det(T[0:2, 0:2]) - 1) < 1e-10:
            return T
        R = T[0:2, 0:2]
        epsilon = 0.001
        if abs(R[0, 0] - R[1, 1]) > epsilon or abs(R[1, 0] + R[0, 1]) > epsilon:
            print("WARNING: this is not a proper rigid transformation:", R)
            return T
        a = (R[0, 0] + R[1, 1]) / 2
        b = (-R[1, 0] + R[0, 1]) / 2
        s = np.sqrt(a**2 + b**2)
        a /= s
        b /= s
        R[0, 0] = a
        R[0, 1] = b
        R[1, 0] = -b
        R[1, 1] = a
        T[0:2, 0:2] = R
    if dim == 3:
        if abs(np.linalg.det(T[0:3, 0:3]) - 1) < 1e-10:
            return T
        c1 = T[0:3, 1]
        c2 = T[0:3, 2]
        c1 /= np.linalg.norm(c1)
        c2 /= np.linalg.norm(c2)
        newcol0 = np.cross(c1, c2)
        newcol1 = np.cross(c2, newcol0)
        T[0:3, 0] = newcol0
        T[0:3, 1] = newcol1
        T[0:3, 2] = c2
    return T


def carrot(xbar):
    """Overloaded operator. converts 3x1 vectors into a member of Lie Alebra so(3)
        Also, converts 6x1 vectors into a member of Lie Algebra se(3)
    Args:
        xbar (np.ndarray): if 3x1, xbar is a vector of rotation angles, if 6x1 a vector of 3 trans and 3 rot angles.
    Returns:
        np.ndarray: Lie Algebra 3x3 matrix so(3) if input 3x1, 4x4 matrix se(3) if input 6x1.
    """
    x = xbar.squeeze()
    if x.shape[0] == 3:
        return np.array([[0, -x[2], x[1]],
                         [x[2], 0, -x[0]],
                         [-x[1], x[0], 0]])
    elif x.shape[0] == 6:
        return np.array([[0, -x[5], x[4], x[0]],
                         [x[5], 0, -x[3], x[1]],
                         [-x[4], x[3], 0, x[2]],
                         [0, 0, 0, 1]])
    print('WARNING: attempted carrot operator on invalid vector shape')
    return xbar


def se3ToSE3(xi):
    """Converts 6x1 vectors representing the Lie Algebra, se(3) into a 4x4 homogeneous transform in SE(3)
        Lie Vector xi = [rho, phi]^T (6 x 1) --> SE(3) T = [C, r; 0 0 0 1] (4 x 4)
    Args:
        xi (np.ndarray): 6x1 vector
    Returns:
        np.ndarray: 4x4 transformation matrix
    """
    T = np.identity(4, dtype=np.float32)
    rho = xi[0:3].reshape(3, 1)
    phibar = xi[3:6].reshape(3, 1)
    phi = np.linalg.norm(phibar)
    R = np.identity(3)
    if phi != 0:
        phibar /= phi  # normalize
        I = np.identity(3)
        R = np.cos(phi) * I + (1 - np.cos(phi)) * phibar @ phibar.T + np.sin(phi) * carrot(phibar)
        J = I * np.sin(phi) / phi + (1 - np.sin(phi) / phi) * phibar @ phibar.T + \
            carrot(phibar) * (1 - np.cos(phi)) / phi
        rho = J @ rho
    T[0:3, 0:3] = R
    T[0:3, 3:] = rho
    return T


def SE3Tose3(T):
    """Converts 4x4 homogeneous transforms in SE(3) to 6x1 vectors representing the Lie Algebra, se(3)
        SE(3) T = [C, r; 0 0 0 1] (4 x 4) --> Lie Vector xi = [rho, phi]^T (6 x 1)
    Args:
        T (np.ndarray): 4x4 transformation matrix
    Returns:
        np.ndarray: 6x1 vector
    """
    R = T[0:3, 0:3]
    evals, evecs = np.linalg.eig(R)
    idx = -1
    for i in range(3):
        if evals[i].real != 0 and evals[i].imag == 0:
            idx = i
            break
    assert(idx != -1)
    abar = evecs[idx].real.reshape(3, 1)
    phi = np.arccos((np.trace(R) - 1) / 2)
    rho = T[0:3, 3:]
    if phi != 0:
        I = np.identity(3)
        J = I * np.sin(phi) / phi + (1 - np.sin(phi) / phi) * abar @ abar.T + \
            carrot(abar) * (1 - np.cos(phi)) / phi
        rho = np.linalg.inv(J) @ rho
    xi = np.zeros((6, 1))
    xi[0:3, 0:] = rho
    xi[3:, 0:] = phi * abar
    return xi


def rotation_error(T):
    """Calculates a single rotation value corresponding to the upper-left 3x3 rotation matrix.
        Uses axis-angle representation to get a single number for rotation
    Args:
        T (np.ndarray): 4x4 transformation matrix T = [C, r; 0 0 0 1]
    Returns:
        float: rotation
    """
    d = 0.5 * (np.trace(T[0:3, 0:3]) - 1)
    return np.arccos(max(min(d, 1.0), -1.0))


def translation_error(T, dim=3):
    """Calculates a euclidean distance corresponding to the translation vector within a 4x4 transform.
    Args:
        T (np.ndarray): 4x4 transformation matrix T = [C, r; 0 0 0 1]
        dim (int): If dim=2 we only use x,y, otherwise we use all dims.
    Returns:
        float: translation distance
    """
    if dim == 2:
        return np.sqrt(T[0, 3]**2 + T[1, 3]**2)
    return np.sqrt(T[0, 3]**2 + T[1, 3]**2 + T[2, 3]**2)


def wrapto2pi(phi):
    """Ensures that the output angle phi is within the interval [0, 2*pi)"""
    if phi < 0:
        return phi + 2 * np.pi * np.ceil(phi / (-2 * np.pi))
    elif phi >= 2 * np.pi:
        return (phi / (2 * np.pi) % 1) * 2 * np.pi
    return phi


def get_time_from_filename(file):
    """Retrieves an epoch time from a file name in seconds"""
    tstr = str(Path(file).stem)
    gpstime = float(tstr)
    timeconvert = 1e-9
    if len(tstr) < 19:
        timeconvert = 10**(-1 * (len(tstr) - 10))
    return gpstime * timeconvert


EARTH_SEMIMAJOR = 6378137.0
EARTH_SEMIMINOR = 6356752.0
EARTH_ECCEN = 0.081819190842622
a = EARTH_SEMIMAJOR
eccSquared = EARTH_ECCEN**2
eccPrimeSquared = (eccSquared) / (1 - eccSquared)
k0 = 0.9996     # scale factor
DEG_TO_RAD = np.pi / 180
RAD_TO_DEG = 1.0 / DEG_TO_RAD


def LLtoUTM(latitude, longitude):
    """Converts a lat-long position into metric UTM coordinates (x-y-z)
    Args:
        latitude (float) in radians
        longitude (float) in radians
    Returns:
        UTMEasting (float): metric position in easting
        UTMNorthing (float): metric position in northing
        zoneNumber (float): UTM zone number
    """
    while longitude < -1 * np.pi:
        longitude += 2 * np.pi
    while longitude >= np.pi:
        longitude -= 2 * np.pi
    longDeg = longitude * RAD_TO_DEG
    latDeg = latitude * RAD_TO_DEG
    zoneNumber = int((longDeg + 180) / 6) + 1
    # +3 puts origin in middle of zone
    longOrigin = (zoneNumber - 1) * 6 - 180 + 3
    longOriginRad = longOrigin * DEG_TO_RAD
    N = a / np.sqrt(1 - eccSquared * np.sin(latitude) * np.sin(latitude))
    T = np.tan(latitude) * np.tan(latitude)
    C = eccPrimeSquared * np.cos(latitude) * np.cos(latitude)
    A = np.cos(latitude) * (longitude - longOriginRad)
    M = a * ((1 - eccSquared / 4 - 3 * eccSquared * eccSquared / 64 - 5 * eccSquared * eccSquared * eccSquared / 256) * latitude -
        (3 * eccSquared / 8 + 3 * eccSquared * eccSquared / 32 + 45 * eccSquared * eccSquared * eccSquared / 1024) * np.sin(2 * latitude) +
        (15 * eccSquared * eccSquared / 256 + 45 * eccSquared * eccSquared * eccSquared / 1024) * np.sin(4 * latitude) -
        (35 * eccSquared * eccSquared * eccSquared / 3072) * np.sin(6 * latitude))
    UTMEasting = k0 * N * (A + (1 - T + C) * A * A * A / 6 +
        (5 - 18 * T + T * T + 72 * C - 58 * eccPrimeSquared) * A * A * A * A * A / 120) + 500000.0
    UTMNorthing = k0 * (M + N * np.tan(latitude) *
        (A * A / 2 + (5 - T + 9 * C + 4 * C * C) * A * A * A * A / 24 +
        (61 - 58 * T + T * T + 600 * C - 330 * eccPrimeSquared) * A * A * A * A * A * A / 720))

    if latitude < 0:
        # 10000000 meter offset for southern hemisphere
        UTMNorthing += 10000000.0

    return UTMEasting, UTMNorthing, zoneNumber


def get_gt_data_for_frame(root, sensType, frame):
    """Retrieves ground truth applanix data for a given sensor frame
    Args:
        root (str): path to the sequence root
        sensType (str): [camera, lidar, or radar]
        frame (str): name/timestampd of the given sensor frame (without the extension)
    Returns:
        gt (list): A list of ground truth values from the applanix sensor_poses.scv
    """
    posepath = osp.join(root, 'applanix', sensType + '_poses.csv')
    with open(posepath, 'r') as f:
        f.readline()  # header
        for line in f:
            if line.split(',')[0] == frame:
                return [float(x) for x in line.split(',')]
    assert(0), 'gt not found for root: {} sensType: {} frame: {}'.format(root, sensType, frame)
    return None


def binaryDistSearch(arr, l, r, x):
    """Performs a binary search to find the index of the element in arr
    which is closest to x. O(log n)
    Note: this function may also return the second-closest element to x,
    so a follow-up check must be done on one index above and below
    Args:
        arr (list): Sorted list of float values
        l (int): index of the left-most section of the array to search
        r (int): index of the right-most section of the array to search
        x (float): query value
    Returns:
        d (float): absolute distance between arr[idx] and x
        idx (int): index of the closest element in array to x
    """
    if r == l + 1:
        if abs(arr[l] - x) < abs(arr[r] - x):
            return abs(arr[l] - x), l
        else:
            return abs(arr[r] - x), r
    mid = l + (r - l) // 2
    if arr[mid] == x:
        return 0, mid
    elif r - l == 0:
        return abs(arr[l] - x), l
    elif arr[mid] > x:
        return binaryDistSearch(arr, l, mid - 1, x)
    else:
        return binaryDistSearch(arr, mid + 1, r, x)


def get_closest_index(query, targets):
    """Retrieves the index of the element in targets that is closest to query O(log n)
    Args:
        query (float): query value
        targets (list): Sorted list of float values
    Returns:
        idx (int): index of the closest element in the array to x
    """
    d, idx = binaryDistSearch(targets, 0, len(targets)-1, query)
    # check if index above or below is closer to query
    if targets[idx] < query and idx < len(targets) - 1:
        if abs(targets[idx + 1] - query) < d:
            return idx + 1
    elif targets[idx] > query and idx > 0:
        if abs(targets[idx - 1] - query) < d:
            return idx - 1
    return idx


def is_sorted(x):
    """Returns True is x is a sorted list, otherwise False"""
    return (np.diff(x) >= 0).all()



import math
import xml.etree.ElementTree as xml
import numpy as np
import pyproj


class Point:
    def __init__(self):
        self.x = None
        self.y = None


class MapframeProjector:
    """
    Class to convert incoming utm or lat/long points to be in a certain frame of reference (x/y origin & rotation).
    **If using lat/long, convention is x is long, y is lat.**
    """
    def __init__(self, x_origin, y_origin, rot_mtx=None, utm=False):
        if not utm:
            self.zone = math.floor((x_origin+180.)/6)+1  # convert to utm zone. doesnt work for all zones
            self.p = pyproj.Proj(proj='utm', ellps='WGS84', zone=self.zone, datum='WGS84')
            [self.x_origin, self.y_origin] = self.p(x_origin, y_origin)
        elif utm:
            self.zone = 17  # Hardcode for now
            self.p = pyproj.Proj(proj='utm', ellps='WGS84', zone=self.zone, datum='WGS84')
            self.x_origin = x_origin
            self.y_origin = y_origin

        if rot_mtx is not None:
            self.rot_mtx = rot_mtx[0:2, 0:2]  # Just take 2d component (x, y) of rotation matrix
        else:
            self.rot_mtx = None

    def proj_latlon(self, lat, lon):
        if self.rot_mtx is not None:
            [x, y] = self.p(lon, lat)
            rotated = np.matmul(self.rot_mtx, np.array([[x-self.x_origin], [y-self.y_origin]]))
            return rotated[0][0], rotated[1][0]
        else:
            [x, y] = self.p(lon, lat)
            return [x - self.x_origin, y - self.y_origin]


def get_way_points(element, point_dict):
    # Get all points in a way in the osm file
    x_pts = []
    y_pts = []
    for nd in element.findall("nd"):
        pt_id = int(nd.get("ref"))
        point = point_dict[pt_id]
        x_pts.append(point.x)
        y_pts.append(point.y)
    return x_pts, y_pts


def draw_map(filename, axes, x_origin, y_origin, rot_mtx=None, utm=False):
    """
    Draws a map from an osm file on the given axis, wrt a given frame.

    Args:
        filename: path to the osm map file to draw
        axes: axes to draw the map on
        x_origin: x coords of origin frame
        y_origin: y coords of origin frame
        rot_mtx: rotation matrix of the map from wrt to an ENU frame
        utm: flag for whether the origin is expressed in UTM or lat/long. If using lat/long, x is long, y is lat
    """
    map_data = xml.parse(filename).getroot()
    projector = MapframeProjector(x_origin, y_origin, rot_mtx, utm)

    point_dict = {}
    for node in map_data.findall("node"):
        point = Point()
        point.x, point.y = projector.proj_latlon(float(node.get('lat')), float(node.get('lon')))
        point_dict[int(node.get('id'))] = point

    axes.patch.set_facecolor('white')
    for way in map_data.findall('way'):  # RN: just plot everything as black
        type_dict = dict(color="black", linewidth=1, zorder=10)
        x_list, y_list = get_way_points(way, point_dict)
        axes.plot(x_list, y_list, **type_dict)



import setuptools

from pathlib import Path

this_directory = Path(__file__).parent
with open(str(this_directory / "README.md"), encoding='utf-8') as f:
	long_description = f.read()

setuptools.setup(
	name='asrl-pyboreas',
	version='1.0.2',
	description='A toolkit for working with the Boreas dataset in Python',
	long_description=long_description,
	long_description_content_type='text/markdown',
	author='Keenan Burnett, Andrew Li, Shichen Lu, Jingxing Qian, Yuchen Wu, David Yoon',
	author_email='boreas@robotics.utias.utoronto.ca',
	url='https://github.com/utiasASRL/boreas-devkit',
	license='BSD',
	packages=setuptools.find_packages(),
	python_requires='>=3.8',
	install_requires=["numpy>=1.21.0",
					  "opencv-python>=4.5.3.56",
					  "matplotlib>=3.4.2",
					  "tqdm>=4.60.0",
					  "pyproj>=3.1.0",
					  "utm>=0.7.0",
					  "asrl-pysteam>=1.0.0"]
)



import os.path as osp
from pathlib import Path
import numpy as np
import cv2
import matplotlib.pyplot as plt
from matplotlib import cm

from pyboreas.data.pointcloud import PointCloud
from pyboreas.utils.utils import get_transform, yawPitchRollToRot, get_time_from_filename, load_lidar
from pyboreas.utils.utils import get_gt_data_for_frame
from pyboreas.utils.radar import load_radar, radar_polar_to_cartesian
from pyboreas.vis.vis_utils import vis_lidar, vis_camera, vis_radar


class Sensor:
    def __init__(self, path):
        self.path = path
        p = Path(path)
        self.frame = p.stem
        self.sensType = p.parts[-2]
        self.seqID = p.parts[-3]
        self.seq_root = str(Path(*p.parts[:-2]))
        self.sensor_root = osp.join(self.seq_root, self.sensType)
        self.pose = np.identity(4, dtype=np.float64)  # T_enu_sensor
        self.velocity = np.zeros((6, 1))   # 6 x 1 velocity in ENU frame [v_se_in_e; w_se_in_e] 
        self.body_rate = np.zeros((6, 1))  # 6 x 1 velocity in sensor frame [v_se_in_s; w_se_in_s]
        self.timestamp = get_time_from_filename(self.frame)

    def init_pose(self, data=None):
        """Initializes pose variables with ground truth applanix data
        Args:
            data (list): A list of floats corresponding to the line from the sensor_pose.csv file
                with the matching timestamp
        """
        if data is not None:
            gt = [float(x) for x in data]
        else:
            gt = get_gt_data_for_frame(self.seq_root, self.sensType, self.frame)
        self.pose = get_transform(gt)
        wbar = np.array([gt[12], gt[11], gt[10]]).reshape(3, 1)
        wbar = np.matmul(self.pose[:3, :3], wbar).squeeze()
        self.velocity = np.array([gt[4], gt[5], gt[6], wbar[0], wbar[1], wbar[2]]).reshape(6, 1)
        vbar = np.array([gt[4], gt[5], gt[6]]).reshape(3, 1)
        vbar = np.matmul(self.pose[:3, :3].T, vbar).squeeze()
        self.body_rate = np.array([vbar[0], vbar[1], vbar[2], gt[12], gt[11], gt[10]]).reshape(6, 1)


class Lidar(Sensor, PointCloud):
    def __init__(self, path):
        Sensor.__init__(self, path)
        self.points = None

    def load_data(self):
        self.points = load_lidar(self.path)
        return self.points

    def visualize(self, **kwargs):
        vis_lidar(self, **kwargs)

    def unload_data(self):
        self.points = None

    # TODO: get_bounding_boxes()
    # TODO: get_semantics()


class Camera(Sensor):
    def __init__(self, path):
        Sensor.__init__(self, path)
        self.img = None

    def load_data(self):
        self.img = cv2.imread(self.path)
        return self.img

    def visualize(self, **kwargs):
        vis_camera(self, **kwargs)

    def unload_data(self):
        self.img = None

    # TODO: get_bounding_boxes() # retrieve from file, cache to class variable
    # TODO: get_semantics() # retrieve from file, cache to class variable


class Radar(Sensor):
    def __init__(self, path):
        Sensor.__init__(self, path)
        self.resolution = 0.0596
        self.timestamps = None
        self.azimuths = None
        self.polar = None
        self.cartesian = None
        self.mask = None

    def load_data(self):
        # Loads polar radar data, timestamps, azimuths, and resolution value
        # Additionally, loads a pre-computed cartesian radar image and binary mask if they exist.
        self.timestamps, self.azimuths, _, self.polar, self.resolution = load_radar(self.path)
        cart_path = osp.join(self.sensor_root, 'cart', self.frame + '.png')
        if osp.exists(cart_path):
            self.cartesian = cv2.imread(cart_path, cv2.IMREAD_GRAYSCALE)
        mask_path = osp.join(self.sensor_root, 'mask', self.frame + '.png')
        if osp.exists(mask_path):
            self.mask = cv2.imread(mask_path, cv2.IMREAD_GRAYSCALE)
        return self.timestamps, self.azimuths, self.polar

    def unload_data(self):
        self.timestamps = None
        self.azimuths = None
        self.polar = None
        self.cartesian = None
        self.mask = None

    def get_cartesian(self, cart_resolution, cart_pixel_width, polar=None, in_place=True):
        """Converts a polar scan from polar to Cartesian format
        Args:
            cart_resolution (float): resolution of the output Cartesian image in (m / pixel)
            cart_pixel_width (int): width of the output Cartesian image in pixels
            polar (np.ndarray): if supplied, this function will use this input and not self.polar.
            in_place (bool): if True, self.cartesian is updated.
        """
        if polar is None:
            polar = self.polar
        cartesian = radar_polar_to_cartesian(self.azimuths, polar, self.resolution,
                                             cart_resolution, cart_pixel_width)
        if in_place:
            self.cartesian = cartesian
        return cartesian

    def visualize(self, **kwargs):
        vis_radar(self, **kwargs)

    # TODO: get_bounding_boxes() # retrieve from file, cache to class variable



# TODO: plot odometry results vs. ground truth

import sys
import glob
import csv

import cv2
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
import numpy as np
from tqdm import tqdm

from pyboreas.vis.boreas_plotter import BoreasPlotter
from pyboreas.data.sequence import Sequence
from pyboreas.utils.utils import get_closest_index

# matplotlib.use("tkagg")  # slu: for testing with ide


def get_closest_frame(query_time, target_times, targets):
    closest = get_closest_index(query_time, target_times)
    assert(np.abs(query_time - target_times[closest]) < 1.0), "closest to query: {} not found.".format(query_time)
    return targets[closest]


class BoreasVisualizer:
    """Main class for loading the Boreas dataset for visualization.

    Loads relevant data, transforms, and labels and provides access to several visualization options.
    Currently only works for one track at a time.
    """

    def __init__(self, sequence):
        """Initialize the class with the corresponding data, transforms and labels.

        Args:
            sequence: Sequence object to base the visualization on
        """
        self.sequence = sequence
        self.calib = sequence.calib
        self.track_length = len(sequence.lidar_frames)
        self.lidar_frames = sequence.lidar_frames
        lstamps = [frame.timestamp for frame in sequence.lidar_frames]
        cstamps = [frame.timestamp for frame in sequence.camera_frames]
        rstamps = [frame.timestamp for frame in sequence.radar_frames]
        # Get corresponding camera and radar frame for each lidar frame
        self.camera_frames = [get_closest_frame(lstamp, cstamps, sequence.camera_frames) for lstamp in lstamps]
        self.radar_frames = [get_closest_frame(lstamp, rstamps, sequence.radar_frames) for lstamp in lstamps]

        # # Load label data
        # print("Loading Labels...", flush=True)
        # with open(self.label_file, 'r') as file:
        #     raw_labels = json.load(file)
        #     for label in tqdm(raw_labels):
        #         self.labels.append(label['cuboids'])

    def visualize(self, frame_idx, predictions=None, mode='both', show=True):
        """
        Visualize the sequence.

        Args:
            frame_idx: the frame index in the current sequence to visualize
            predictions: user generated predictions to also be visualized
            mode: 'persp' for perspective (camera) view, 'bev' for top down view, or 'both'
            show: flag to show the matplotlib plot. Should only be off for exporting to video

        Returns: the BoreasPlotter object used for visualizing

        """
        boreas_plot = BoreasPlotter(self, frame_idx, mode=mode)
        boreas_plot.update_plots(frame_idx)
        if show:
            plt.show()
        else:
            plt.close(boreas_plot.fig)

        return boreas_plot

    def export_vis_video(self, name, mode='both'):
        """
        Exports the current sequence visualization to video. Defaults to current working directory for save location.

        Args:
            name: name for the exported video file (don't include extension)
            mode: 'persp' for perspective (camera) view, 'bev' for top down view, or 'both'
        """
        imgs = []
        # Render the matplotlib figs to images
        print("Exporting Visualization to Video")
        for i in tqdm(range(len(self.lidar_frames)), file=sys.stdout):
            bplot = self.visualize(frame_idx=i, mode=mode, show=False)
            canvas = FigureCanvas(bplot.fig)
            canvas.draw()
            graph_image = np.array(bplot.fig.canvas.get_renderer()._renderer)
            graph_image = cv2.cvtColor(graph_image, cv2.COLOR_RGB2BGR)
            imgs.append(graph_image)

        # Write the images to video
        # MJPG encoder is part of cv2
        out = cv2.VideoWriter(name + ".avi", cv2.VideoWriter_fourcc(*'MJPG'), 15, (graph_image.shape[1], graph_image.shape[0]))
        for i in range(len(imgs)):
            out.write(imgs[i])
        out.release()


if __name__ == '__main__':
    sequence = Sequence("/home/shichen/datasets/", ["boreas_mini_v2", 1616518050000000, 1616518060000000])
    dataset = BoreasVisualizer(sequence)
    # for name in ["both", "persp", "bev"]:
    #     dataset.export_vis_video(name, name)
    dataset.visualize(50)



import argparse
from pyboreas.utils.odometry import get_sequences, get_sequence_poses, compute_kitti_metrics

if __name__ == '__main__':
    # parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('--pred', default='test/demo/pred', type=str, help='path to prediction files')
    parser.add_argument('--gt', default='test/demo/gt', type=str, help='path to groundtruth files')
    args = parser.parse_args()

    # parse sequences
    seq = get_sequences(args.pred, '.txt')
    T_pred, times_pred, seq_lens_pred = get_sequence_poses(args.pred, seq)
    T_gt, times_gt, seq_lens_gt = get_sequence_poses(args.gt, seq)

    # compute errors
    t_err, r_err = compute_kitti_metrics(T_gt, T_pred, times_gt, times_pred,
                                         seq_lens_gt, seq_lens_pred, seq, args.pred, step_size=10)

    # print out results
    print('Evaluated sequences: ', seq)
    print('Overall error: ', t_err, ' %, ', r_err, ' deg/m')